# ovirt-imageio
# Copyright (C) 2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


class Range:

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __lt__(self, other):
        return (self.start, self.end) < (other.start, other.end)

    def __len__(self):
        return self.end - self.start

    def __eq__(self, other):
        return (self.__class__ == other.__class__ and
                self.start == other.start and
                self.end == other.end)

    def __ne__(self, other):
        return not self == other


def merge_ranges(ranges):
    """
    Gets an iterable of Range objects and returns
    a sorted list of the merged ranges.
    """
    ranges = sorted(ranges)
    if not ranges:
        return ranges
    merged = [ranges.pop(0)]
    for range in ranges:
        current = merged[-1]
        if current.end >= range.start:
            current.end = max(current.end, range.end)
        else:
            merged.append(range)
    return merged



# ovirt-imageio
# Copyright (C) 2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import pytest

from ovirt_imageio._internal import measure
from ovirt_imageio._internal.measure import Range


class TestRange:

    """
    To make the cases visual, each test case is described using two ranges:
    - The range that change in each test case is drawn as follows: "*****".
    - The range that we compare the others to is drawn as follows: "-----".
    """
    @pytest.mark.parametrize("start,end,expected", [
        #          *****
        #     -----
        (15, 20, True),

        #         *****
        #     -----
        (10, 20, True),

        #         *
        #     -----
        (10, 10, True),

        #       *****
        #     -----
        (7, 20, True),

        #       ***
        #     -----
        (7, 10, True),

        #      ***
        #     -----
        (7, 8, True),

        #     *******
        #     -----
        (5, 15, True),

        #     *****
        #     -----
        (5, 10, False),

        #     ***
        #     -----
        (5, 8, False),

        #     *
        #     -----
        (5, 5, False),

        #   *********
        #     -----
        (3, 15, False),

        #   *******
        #     -----
        (3, 10, False),

        # *****
        #     -----
        (3, 5, False),

        #   *****
        #     -----
        (3, 8, False),

        # ***
        #     -----
        (3, 4, False),
    ])
    def test_lt(self, start, end, expected):
        assert (Range(5, 10) < Range(start, end)) == expected

    def test_eq(self):
        assert Range(5, 10) == Range(5, 10)

    def test_ne(self):
        assert Range(5, 10) != Range(15, 20)


@pytest.mark.parametrize("orig_ranges,merged_ranges", [
    # No range
    ([], []),

    # Single range
    ([(5, 10)], [(5, 10)]),

    # Two consecutive ranges.
    ([(5, 10), (10, 20)], [(5, 20)]),

    # Two ranges with a "hole" in between.
    ([(5, 10), (15, 20)], [(5, 10), (15, 20)]),

    # Two overlapping ranges.
    ([(5, 10), (5, 20)], [(5, 20)]),

    # Three unsorted and partially overlapping ranges.
    ([(5, 10), (0, 3), (7, 20)],
     [(0, 3), (5, 20)]),

    # Three separate ranges.
    ([(5, 10), (15, 20), (25, 30)],
     [(5, 10), (15, 20), (25, 30)]),

    # Three unsorted and overlapping ranges.
    ([(5, 10), (0, 5), (10, 20)], [(0, 20)]),
])
def test_merge_ranges(orig_ranges, merged_ranges):
    orig_ranges = [Range(start, end) for start, end in orig_ranges]
    merged_ranges = [Range(start, end) for start, end in merged_ranges]
    assert measure.merge_ranges(orig_ranges) == merged_ranges



# ovirt-imageio
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from collections import namedtuple
from functools import partial

from .. import errors
from .. import util

from . import file
from . import http
from . import nbd

_modules = {
    "file": file,
    "nbd": nbd,
    "https": http,
}


class Unsupported(Exception):
    """ Requested backend is not supported """


class Context(namedtuple("Context", "backend,buffer")):
    """
    Backend context stored per ticket connection.
    """
    __slots__ = ()

    def close(self):
        try:
            self.backend.close()
        finally:
            self.buffer.close()


class Closer:
    """
    Call function when closed.
    """
    def __init__(self, func):
        self.close = func


def supports(name):
    return name in _modules


def get(req, ticket, config):
    """
    Return a connection backend for this ticket.

    On the first call, open the backend and cache it in the connection context.
    The backend will be closed when the connection is closed.

    Thread safety: requests are accessed by the single connection thread, no
    locking is needed.
    """
    try:
        return ticket.get_context(req.connection_id)
    except KeyError:
        if not supports(ticket.url.scheme):
            raise Unsupported(
                "Unsupported backend {!r}".format(ticket.url.scheme))

        mode = "r+" if "write" in ticket.ops else "r"
        module = _modules[ticket.url.scheme]

        # If HTTP backend has no explict CA file configuration, use CA file
        # from TLS configuration.
        ca_file = config.backend_http.ca_file or config.tls.ca_file

        backend = module.open(
            ticket.url,
            mode=mode,
            sparse=ticket.sparse,
            dirty=ticket.dirty,
            max_connections=config.daemon.max_connections,
            cafile=ca_file)

        backend_config = getattr(config, "backend_" + backend.name)
        buf = util.aligned_buffer(backend_config.buffer_size)
        ctx = Context(backend, buf)

        # Keep the context in the ticket so we monitor the number of
        # connections using the ticket.
        try:
            ticket.add_context(req.connection_id, ctx)
        except errors.AuthorizationError:
            # The ticket was canceled. close the backend and let the caller
            # handle the failure.
            ctx.close()
            raise

        # Register a closer removing the context when the connection is closed.
        req.context[ticket.uuid] = Closer(
            partial(ticket.remove_context, req.connection_id))

        return ctx



# ovirt-imageio
# Copyright (C) 2015-2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from distutils.core import setup
from distutils.core import Extension

from ovirt_imageio._internal import version

setup(
    author="oVirt Authors",
    author_email="devel@ovirt.org",
    description="oVirt imageio",
    license="GNU GPLv2+",
    name="ovirt-imageio",
    packages=[
        "ovirt_imageio",
        "ovirt_imageio._internal",
        "ovirt_imageio._internal.backends",
        "ovirt_imageio.client",
    ],
    platforms=["Linux"],
    scripts=["ovirt-imageio"],
    url="https://gerrit.ovirt.org/ovirt-imageio",
    version=version.string,
    ext_modules=[
        Extension(
            "ovirt_imageio/_internal/ioutil",
            sources=["ovirt_imageio/_internal/ioutil.c"]),
    ]
)



# ovirt-imageio
# Copyright (C) 2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import hashlib
import logging

from . import backends
from . import blkhash
from . import errors
from . import http
from . import ioutil
from . import ops
from . import util
from . import validate

log = logging.getLogger("checksum")

ALGORITHMS = frozenset(hashlib.algorithms_available)

# Limit allowed block size to avoid abusing server resources.
MIN_BLOCK_SIZE = blkhash.BLOCK_SIZE // 4
MAX_BLOCK_SIZE = blkhash.BLOCK_SIZE * 4


class Checksum:
    """
    Handle requests for the /images/ticket-id/checksum resource.
    """

    def __init__(self, config, auth):
        self.config = config
        self.auth = auth

    def get(self, req, resp, ticket_id):
        if not ticket_id:
            raise http.Error(http.BAD_REQUEST, "Ticket id is required")

        algorithm = validate.enum(
            req.query,
            "algorithm",
            ALGORITHMS,
            default=blkhash.ALGORITHM)

        try:
            block_size = int(req.query.get("block_size", blkhash.BLOCK_SIZE))
        except ValueError:
            raise http.Error(
                http.BAD_REQUEST,
                "Invalid block size: {!r}".format(req.query["block_size"]))

        if not MIN_BLOCK_SIZE <= block_size <= MAX_BLOCK_SIZE:
            raise http.Error(
                http.BAD_REQUEST,
                "Block size out of allowed range: {}-{}"
                .format(MIN_BLOCK_SIZE, MAX_BLOCK_SIZE))

        if block_size % 4096:
            raise http.Error(
                http.BAD_REQUEST, "Block size is not aligned to 4096")

        try:
            ticket = self.auth.authorize(ticket_id, "read")
            ctx = backends.get(req, ticket, self.config)
        except errors.AuthorizationError as e:
            raise http.Error(http.FORBIDDEN, str(e))

        log.info("[%s] CHECKSUM ticket=%s algorithm=%s block_size=%s",
                 req.client_addr, ticket_id, algorithm, block_size)

        # For simplicity we create a new buffer even if block_size is same as
        # ctx.buffer length.

        with util.aligned_buffer(block_size) as buf:
            op = Operation(ctx.backend, buf, algorithm, clock=req.clock)
            try:
                checksum = ticket.run(op)
            except errors.AuthorizationError as e:
                resp.close_connection()
                raise http.Error(http.FORBIDDEN, str(e)) from None

        resp.send_json(checksum)


class Algorithms:
    """
    Handle requests for the /images/ticket-id/checksum/algorithms resource.
    """

    def __init__(self, config, auth):
        self.config = config
        self.auth = auth

    def get(self, req, resp, *args):
        # Server information, no authorization needed.
        resp.send_json({"algorithms": sorted(ALGORITHMS)})


class Operation(ops.Operation):
    """
    Checksum operation.
    """

    name = "checksum"

    def __init__(self, backend, buf, algorithm, detect_zeroes=True,
                 clock=None):
        super().__init__(size=backend.size(), buf=buf, clock=clock)
        self._backend = backend
        self._algorithm = algorithm
        self._detect_zeroes = detect_zeroes

    def _run(self):
        block_size = len(self._buf)
        # Only blakse2b and blake2s support variable digest size, and 32 works
        # with both and is large enough.
        if self._algorithm.startswith("blake2"):
            digest_size = blkhash.DIGEST_SIZE
        else:
            digest_size = None

        h = blkhash.Hash(
            block_size=block_size,
            algorithm=self._algorithm,
            digest_size=digest_size)

        for block in blkhash.split(self._backend.extents("zero"), block_size):
            if block.zero:
                h.zero(block.length)
            else:
                with memoryview(self._buf)[:block.length] as view:
                    self._backend.seek(block.start)
                    self._backend.readinto(view)
                    if self._detect_zeroes and ioutil.is_zero(view):
                        h.zero(block.length)
                    else:
                        h.update(view)

            self._done += block.length

            if self._canceled:
                raise ops.Canceled

        return {
            "algorithm": self._algorithm,
            "block_size": block_size,
            "checksum": h.hexdigest(),
        }


def compute(backend, buf, algorithm=blkhash.ALGORITHM, detect_zeroes=True):
    """
    Compute image checksum.
    """
    op = Operation(backend, buf, algorithm, detect_zeroes=detect_zeroes)
    return op.run()



# ovirt-imageio
# Copyright (C) 2017 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import time

import pytest

from ovirt_imageio._internal import config
from ovirt_imageio._internal import errors
from ovirt_imageio._internal import ops
from ovirt_imageio._internal import util
from ovirt_imageio._internal.auth import Ticket, Authorizer

from test import testutil

CHUNK_SIZE = 8 * 1024**2


class Operation:
    """
    Used to fake a ops.Operation object.
    """

    def __init__(self, offset=0, size=0):
        self.offset = offset
        self.size = size
        self.done = 0
        self.canceled = False

    def run(self):
        if self.canceled:
            raise ops.Canceled
        self.done = self.size

    def cancel(self):
        self.canceled = True


def test_transfered_nothing():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    assert ticket.transferred() == 0


def test_transfered_inactive_empty_ops():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    ticket.run(Operation(0, 0))
    assert ticket.transferred() == 0

    ticket.run(Operation(1000, 0))
    assert ticket.transferred() == 0


def test_transfered_inactive_ordered_ops():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    ticket.run(Operation(0, 100))
    assert ticket.transferred() == 100

    ticket.run(Operation(100, 100))
    assert ticket.transferred() == 200

    ticket.run(Operation(200, 100))
    assert ticket.transferred() == 300


def test_transfered_inactive_unordered_ops():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    ticket.run(Operation(100, 100))
    assert ticket.transferred() == 100

    ticket.run(Operation(0, 100))
    assert ticket.transferred() == 200

    ticket.run(Operation(200, 100))
    assert ticket.transferred() == 300


def test_transfered_inactive_overlapping_ops():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    ticket.run(Operation(0, 120))
    assert ticket.transferred() == 120

    ticket.run(Operation(100, 100))
    assert ticket.transferred() == 200

    ticket.run(Operation(180, 120))
    assert ticket.transferred() == 300


def test_transfered_inactive_non_continuous_ops():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    # Run 2 non-continutes operations
    ticket.run(Operation(0, 100))
    ticket.run(Operation(200, 100))
    assert ticket.transferred() == 200

    # Run last operation filling the hole - with some overlap.
    ticket.run(Operation(80, 120))
    assert ticket.transferred() == 300


def test_transfered_ongoing_concurrent_ops():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))

    # Start 2 ongoing operations:
    # ongoing: 0-0, 100-100
    # completed:
    op1 = Operation(0, 100)
    ticket._add_operation(op1)
    assert ticket.transferred() == 0
    assert ticket.active()

    op2 = Operation(100, 100)
    ticket._add_operation(op2)
    assert ticket.transferred() == 0
    assert ticket.active()

    # Consume op1 data:
    # ongoing: 0-100, 100-100
    # completed:
    op1.run()
    ticket._remove_operation(op1)
    assert ticket.transferred() == 100
    assert ticket.active()

    # Consume op2 data:
    # ongoing: 0-100, 100-200
    # completed:
    op2.run()
    ticket._remove_operation(op2)
    assert ticket.transferred() == 200
    assert not ticket.active()


def test_transfered_ongoing_overlapping_ops():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))

    # Start 2 ongoing operations.
    # ongoing: 0-0, 80-80
    # completed:
    op1 = Operation(0, 120)
    op2 = Operation(80, 120)
    ticket._add_operation(op1)
    ticket._add_operation(op2)
    assert ticket.transferred() == 0
    assert ticket.active()

    # Consume op1 data:
    # ongoing: 0-120, 80-80
    # completed:
    op1.run()
    ticket._remove_operation(op1)
    assert ticket.transferred() == 120
    assert ticket.active()

    # Consume op2 data:
    # ongoing: 0-120, 80-200
    # completed:
    op2.run()
    ticket._remove_operation(op2)
    assert ticket.transferred() == 200
    assert not ticket.active()


def test_transfered_ongoing_non_continues_ops():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))

    # Start 2 ongoing operations.
    # ongoing: 0-0, 200-200
    # completed:
    op1 = Operation(0, 100)
    op2 = Operation(200, 100)
    ticket._add_operation(op1)
    ticket._add_operation(op2)
    assert ticket.transferred() == 0
    assert ticket.active()

    # Consume op1 data:
    # ongoing: 0-100, 200-200
    # completed:
    op1.run()
    ticket._remove_operation(op1)
    assert ticket.transferred() == 100

    # Consume op2 data:
    # ongoing: 0-100, 200-300
    # completed:
    op2.run()
    ticket._remove_operation(op2)
    assert ticket.transferred() == 200


@pytest.mark.benchmark
def test_run_operation_benchmark():
    # Run 1000000 operations with 4 concurrent threads.
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    operations = 10**6
    workers = 4
    chunk = 10**9
    step = chunk * workers // operations

    def worker(offset, size):
        while offset < size:
            ticket.run(Operation(offset, step))
            offset += step

    start = time.monotonic()

    threads = []
    try:
        for i in range(workers):
            t = util.start_thread(worker, args=(i * chunk, chunk))
            threads.append(t)
    finally:
        for t in threads:
            t.join()

    elapsed = time.monotonic() - start

    print("%d operations, %d concurrent threads in %.3f seconds (%d nsec/op)"
          % (operations, workers, elapsed, elapsed * 10**9 // operations))


@pytest.mark.benchmark
@pytest.mark.parametrize("concurrent", [1, 2, 4, 8])
def test_transferred_benchmark(concurrent):
    # Time trransferred call with multiple ongoing and completed operations.
    ticket = Ticket(testutil.create_ticket(ops=["read"]))

    calls = 10000

    # Add some completed ranges - assume worst case when ranges are not
    # continues.
    for i in range(concurrent):
        ticket.run(Operation(i * 1000, 100))

    # Add some ongoing operations - assume worst case when ranges are not
    # continues.
    for i in range(concurrent):
        ticket._add_operation(Operation(i * 1000 + 200, 100))

    # Time transferred call - merging ongoing and completed ranges.
    start = time.monotonic()
    for i in range(calls):
        ticket.transferred()
    elapsed = time.monotonic() - start

    print("%d concurrent operations, %d calls in %.3f seconds (%d nsec/op)"
          % (concurrent, calls, elapsed, elapsed * 10**9 // calls))


@pytest.mark.parametrize("arg", [
    "not a dict",
    ["not", "a", "dict"],
    1,
    3.1,
    True,
    False,
    None
])
def test_invalid_argument(arg):
    with pytest.raises(errors.InvalidTicket):
        Ticket(arg)


@pytest.mark.parametrize("kw", [
    {"uuid": 1},
    {"size": "not an int"},
    {"ops": "not a list"},
    {"timeout": "not an int"},
    {"url": 1},
    {"transfer_id": 1},
    {"filename": 1},
    {"sparse": 1},
    {"dirty": 1},
])
def test_invalid_parameter(kw):
    with pytest.raises(errors.InvalidTicketParameter):
        Ticket(testutil.create_ticket(**kw))


def test_sparse_unset():
    ticket = Ticket(testutil.create_ticket())
    assert not ticket.sparse


def test_sparse():
    ticket = Ticket(testutil.create_ticket(sparse=True))
    assert ticket.sparse


def test_dirty_unset():
    ticket = Ticket(testutil.create_ticket())
    assert not ticket.dirty


def test_dirty():
    ticket = Ticket(testutil.create_ticket(dirty=True))
    assert ticket.dirty


def test_transfer_id_unset():
    ticket = Ticket(testutil.create_ticket())
    assert ticket.transfer_id is None


def test_transfer_id():
    ticket = Ticket(testutil.create_ticket(transfer_id="123"))
    assert ticket.transfer_id == "123"


def test_repr():
    ticket = Ticket(testutil.create_ticket(
        ops=["read"], filename="tmp_file"))
    ticket_repr = repr(ticket)

    info = ticket.info()
    del info["timeout"]

    for key, value in info.items():
        pair = "%s=%r" % (key, value)
        assert pair in ticket_repr


def test_ticket_run():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    op = Operation(0, 100)
    assert ticket.transferred() == op.done
    assert op.done == 0

    ticket.run(op)

    assert ticket.transferred() == op.done
    assert op.done == 100


def test_cancel_unused():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    ticket.cancel()

    # Ticket is canceled and can be removed immediately.
    assert ticket.canceled
    info = ticket.info()
    assert info["canceled"]
    assert info["connections"] == 0


def test_cancel_timeout():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    ticket.add_context(1, None)

    # Canceling will time out.
    with pytest.raises(errors.TicketCancelTimeout):
        ticket.cancel(timeout=0.001)

    # But ticket is marked as canceled.
    assert ticket.canceled

    # Caller can poll ticket status and remove the ticket when number of
    # connections is zero.
    info = ticket.info()
    assert info["canceled"]
    assert info["connections"] == 1


def test_cancel_async():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    ticket.add_context(1, None)
    ticket.cancel(timeout=0)

    # Ticket is canceled, but it cannot be removed.
    assert ticket.canceled
    info = ticket.info()
    assert info["canceled"]
    assert info["connections"] == 1


def test_cancel_ongoing_operations():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))

    # Few connections are using this ticket. Each running an operation.
    ops = []
    for i in range(4):
        ctx = Context()
        op = Operation()
        ticket.add_context(i, ctx)
        ticket._add_operation(op)
        ops.append(op)

    ticket.cancel(timeout=0)

    # All ongoing operations are canceled.
    assert all(op.canceled for op in ops)


class Context:

    def __init__(self):
        self.closed = False

    def close(self):
        self.closed = True


def test_cancel_wait():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))

    # Add connections using this ticket.
    connections = []
    for i in range(4):
        ctx = Context()
        ticket.add_context(i, ctx)
        connections.append(ctx)

    def close_connections():
        time.sleep(0.1)
        for i in range(4):
            ticket.remove_context(i)

    info = ticket.info()
    assert not info["canceled"]
    assert info["connections"] == 4

    t = util.start_thread(close_connections)
    try:
        ticket.cancel(timeout=10)

        # After the ticket was canceled, number of connections must be zero.
        info = ticket.info()
        assert info["connections"] == 0

        # And all contexts must be closed.
        assert all(ctx.closed for ctx in connections)
    finally:
        t.join()


def test_canceled_fail_run_before():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    ticket.cancel()

    op = Operation()

    # Running operations must fail.
    with pytest.raises(errors.AuthorizationError):
        ticket.run(op)

    # Operation was not run.
    assert op.done == 0


def test_canceled_fail_run_after():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))

    class Operation:

        def __init__(self):
            self.done = False
            self.canceled = False

        def run(self):
            self.done = True
            ticket.cancel()

        def cancel(self):
            self.canceled = True

    op = Operation()

    # If ticket was canceled while ongoing operations are running, ticket run
    # will fail removing the operations.
    with pytest.raises(errors.AuthorizationError):
        ticket.run(op)

    assert op.done


def test_canceled_fail_add_context():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    ticket.cancel()

    ctx = Context()

    # Adding new context must fail.
    with pytest.raises(errors.AuthorizationError):
        ticket.add_context(2, ctx)


def test_get_context_missing():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    with pytest.raises(KeyError):
        ticket.get_context(1)


def test_get_context():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    ctx = Context()
    ticket.add_context(1, ctx)
    assert ticket.get_context(1) is ctx


def test_remove_context_missing():
    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    ticket.add_context(1, Context())
    assert ticket.info()["connections"] == 1

    ticket.remove_context(2)
    assert ticket.info()["connections"] == 1


def test_remove_context_error():

    class FailingContext:

        def __init__(self):
            self.count = 1
            self.closed = False

        def close(self):
            if self.count > 0:
                self.count -= 1
                raise RuntimeError("Cannot close yet")
            self.closed = True

    ticket = Ticket(testutil.create_ticket(ops=["read"]))
    ticket.add_context(1, FailingContext())

    ticket.cancel(timeout=0)

    # If closing a context fails, keep it. The ticket cannot be removed until
    # this context is removed successfully.
    with pytest.raises(RuntimeError):
        ticket.remove_context(1)

    info = ticket.info()
    assert info["connections"] == 1

    # Calling again will close and remove the context successfully.
    ticket.remove_context(1)

    info = ticket.info()
    assert info["connections"] == 0


def test_authorizer_add():
    cfg = config.load(["test/conf.d/daemon.conf"])
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=["read"])
    auth.add(ticket_info)

    ticket = auth.get(ticket_info["uuid"])
    assert ticket.uuid == ticket_info["uuid"]


def test_authorizer_remove_unused():
    cfg = config.load(["test/conf.d/daemon.conf"])
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=["read"])
    auth.add(ticket_info)

    # Ticket is unused so it will be removed.
    auth.remove(ticket_info["uuid"])
    with pytest.raises(KeyError):
        auth.get(ticket_info["uuid"])


def test_authorizer_remove_timeout():
    cfg = config.load(["test/conf.d/daemon.conf"])
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=["read"])
    auth.add(ticket_info)

    ticket = auth.get(ticket_info["uuid"])
    ticket.add_context(1, Context())
    assert ticket.info()["connections"] == 1

    # Use short timeout to keep the tests fast.
    cfg.control.remove_timeout = 0.001

    # Ticket cannot be removed since it is used by connection 1.
    with pytest.raises(errors.TicketCancelTimeout):
        auth.remove(ticket.uuid)

    # Ticket was not removed.
    assert auth.get(ticket.uuid) is ticket

    # The connection was closed, the ticket can be removed now.
    ticket.remove_context(1)
    assert ticket.info()["connections"] == 0

    auth.remove(ticket.uuid)

    # Ticket was removed.
    with pytest.raises(KeyError):
        auth.get(ticket.uuid)


def test_authorizer_remove_async():
    cfg = config.load(["test/conf.d/daemon.conf"])
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=["read"])
    auth.add(ticket_info)

    ticket = auth.get(ticket_info["uuid"])
    ticket.add_context(1, Context())
    assert ticket.info()["connections"] == 1

    # Disable the timeout, so removing a ticket cancel the ticket without
    # waiting, and requiring polling the ticket status.
    cfg.control.remove_timeout = 0

    # Ticket is canceled, but not removed.
    auth.remove(ticket.uuid)
    assert ticket.canceled
    assert ticket.info()["connections"] == 1

    # Ticket was not removed.
    assert auth.get(ticket.uuid) is ticket

    # The connection was closed, the ticket can be removed now.
    ticket.remove_context(1)
    assert ticket.info()["connections"] == 0

    auth.remove(ticket.uuid)

    # Ticket was removed.
    with pytest.raises(KeyError):
        auth.get(ticket.uuid)


def test_authorizer_remove_mising():
    cfg = config.load(["test/conf.d/daemon.conf"])
    auth = Authorizer(cfg)
    # Removing missing ticket does not raise.
    auth.remove("no-such-ticket")


def test_authorize_read():
    cfg = config.load(["test/conf.d/daemon.conf"])
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=["read"])
    auth.add(ticket_info)

    ticket = auth.get(ticket_info["uuid"])
    assert auth.authorize(ticket.uuid, "read") == ticket

    with pytest.raises(errors.AuthorizationError):
        auth.authorize(ticket.uuid, "write")


def test_authorize_write():
    cfg = config.load(["test/conf.d/daemon.conf"])
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=["write"])
    auth.add(ticket_info)

    ticket = auth.get(ticket_info["uuid"])
    assert auth.authorize(ticket.uuid, "write") == ticket

    # "write" implies also "read".
    assert auth.authorize(ticket.uuid, "read") == ticket


def test_authorizer_no_ticket():
    cfg = config.load(["test/conf.d/daemon.conf"])
    auth = Authorizer(cfg)
    with pytest.raises(errors.AuthorizationError):
        auth.authorize("no-such-ticket", "read")


@pytest.mark.parametrize("ops,allowed", [
    (["read"], ["read"]),
    (["write"], ["read", "write"]),
])
def test_authorizer_canceled(ops, allowed):
    cfg = config.load(["test/conf.d/daemon.conf"])
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=ops)
    auth.add(ticket_info)
    ticket = auth.get(ticket_info["uuid"])

    # Cancelling the ticket disables any operation.
    ticket.cancel()

    for op in allowed:
        with pytest.raises(errors.AuthorizationError):
            auth.authorize(ticket.uuid, op)


def test_authorizer_expired():
    cfg = config.load(["test/conf.d/daemon.conf"])
    auth = Authorizer(cfg)
    ticket_info = testutil.create_ticket(ops=["write"])
    auth.add(ticket_info)
    ticket = auth.get(ticket_info["uuid"])

    # Extending with zero timeout expire the ticket.
    ticket.extend(0)

    for op in ("read", "write"):
        with pytest.raises(errors.AuthorizationError):
            auth.authorize(ticket.uuid, op)
